import pytest
from sbx_word_prediction_kb_bert import (
    MODELS,
    TopKPredictor,
)
from transformers import (  # type: ignore [import-untyped]
    BertTokenizer,
    BertForMaskedLM,
)


@pytest.fixture(scope="session")
def kb_bert_predictor() -> TopKPredictor:
    tokenizer_name, tokenizer_revision = MODELS["kb-bert"].tokenizer_name_and_revision()
    tokenizer = BertTokenizer.from_pretrained(
        tokenizer_name, revision=tokenizer_revision
    )
    model = BertForMaskedLM.from_pretrained(
        MODELS["kb-bert"].model_name, revision=MODELS["kb-bert"].model_revision
    )
    return TopKPredictor(model=model, tokenizer=tokenizer)



from itertools import islice
from typing import Tuple
import pytest
from sbx_word_prediction_kb_bert.predictor import TopKPredictor


TEXTS: dict[str, str] = {
    "short": "[MASK] åt glassen utanför kiosken .",
    "long-mask-middle": """
    Frisörbranschen påverkas väldigt negativt av den nuvarande situationen när kunderna inte kan komma till salongerna, vilket förstås gör att många frisörer är hotade av konkurs
    Totalt samlade lokalavdelningarna in 89 563 kronor för aktiviteter riktade till ungdomar under det första halvåret 2015, vilket var en ökning med 39 procent jämfört med samma period föregående år
    Utvärderingen visar att vi kan välja mellan att köpa in nya datorer som är mer kraftfulla och därför klarar de nya systemkraven eller att uppgradera våra datorer med nya grafikkort eller avstå från att installera den nya mjukvaran och
    Sammantaget bedömer vi att det kommer att vara [MASK] att genomföra projektet med befintlig personal och befintliga tekniska resurser och efter projektet kan vi starta förvaltningen av det system som projektet ska leda fram till
    Eftersom dessa rapporter är det huvudsakliga underlaget och empiriska materialet för utvärderingen som vi har genomfört har vi valt att i detalj beskriva hur utvärderingarna kommer till, vilka som genomför dem och hur deras slutrapporter ser ut
    Språkkonsulterna är en byrå av experter inom skriftlig och muntlig kommu­nikation – särskilt utbildade i att anpassa språket till den tänkta målgruppen
    Om jag bestämmer mig för att skriva en lång text så blir jag hungrig. Jag måste laga kaffe och kanske springa ner till bageriet, eller leta efter en karamell eller något annat sött. Och så när jag är klar med mina förberedelse ringer min före detta svärdotter. Hon har just blivit opererad. Det var blindtarmen och något mer som de plockade ut. Hennes mamma säger att hon gnäller och hennes nuvarande svärmorsa förstår inga europeiska språk, inte ett enda ord. Keine wörte bitte! Då sitter man där och ska vara snäll. Den nya svärdottern, hon som efterträdde den gamla, ringer aldrig. Kanske därför tar jag den gamla som en kompensation, eller en evig ursäkt för att inte skriva en lång text.
    Om jag nu bestämmer mig för att skriva en kort text, vad då? Då måste jag absolut börja med att kolla facebook, sedan mailboxen och sedan hur det går där borta i Sverige för de stackars bönderna som söker fruar men har så vansinnigt flotta och fina traktorer och hus som förr i världen skulle betraktas som herrgårdar, och så ringer telefonen. Det är en barndomsvän. Han brukar ringa varannan dag. Han har en ganska gedigen utbildning. Matematiker och gudarna vet vad, och sen fick han hjärnblödning och blev överkörd av en buss och ramlade omkull i Nationalteaterns trappuppgång, i den ordningen. Han orar sig för spionerna. Det finns tydligen spioner överallt. Det finns spioner utanför den amerikanska ambassaden och det finns spioner inni i den amerikanska ambassaden. Nu råkar denna stora vita byggnad, som hyser de amerikanska diplomaterna finnas här alldeles i närheten. De där stackarna som står givakt i alla väder kan åtminstone inte vara spionerna, men spionerna är kanske osynliga? I alla fall har jag fått reda på, genom en konstnär som spelar jultomte året om, att spionerna bakom den norska ambassaden är osynliga men om man står helt stilla kan man höra när de jojkar till varandra.
    Jag går och ser en föreställning och måste naturligtvis skriva lika långt varje gång. I går såg jag en ung man som åt upp ett päron, det var alltså för att vara orginell, för äpplet finns ju som en klassisk föreställning. Denna unga man som har långa ben, vansinnigt långa ben, ovanligt långa ben, lyckades tråka ihjäl femtiotvå personer som sprang ut ur en gammal fiskfabrik och kommer nog aldrig att se en s k föreställning där han finns med nån mer gång. Oj vad det är svårt att konsentrera sig om man har gäster. De ska ha kaffe och de ska underhållas, då kan man heller inte skriva en lång eller en kort text.
    """,
}


def test_short_text(kb_bert_predictor: TopKPredictor) -> None:
    text = "[MASK] åt glassen utanför kiosken ."

    actual = kb_bert_predictor.get_top_k_predictions(text)

    actual = remove_scores(actual)
    expected = "|Vi|Jag|De|Hon|Han|"

    assert actual == expected


def test_rounding(kb_bert_predictor: TopKPredictor) -> None:
    text = "namnet på det hus där historien börjar och slutar [MASK] men annars pratas det mest om huset . "
    kb_bert_predictor.num_decimals = 2
    actual = kb_bert_predictor.get_top_k_predictions(text)
    kb_bert_predictor.num_decimals = 3
    print(f"{actual=}")
    expected = [",:0.92", "-:0.06"]

    num_bars = actual.count("|")
    assert list(islice(actual.split("|"), 1, num_bars)) == expected


def remove_scores(actual):
    return "|".join(map(lambda x: x.split(":")[0], actual.split("|")))


def test_long_text(kb_bert_predictor: TopKPredictor) -> None:
    text = TEXTS["long-mask-middle"]

    assert len(text.split(" ")) > 512
    actual = kb_bert_predictor.get_top_k_predictions(text)

    actual = remove_scores(actual)
    expected = "|möjligt|lönsamt|lättare|fördelaktigt|lätt|"

    assert actual == expected


@pytest.mark.parametrize(
    "texts_key, expected", [("short", (0, 35)), ("long-mask-middle", (462, 864))]
)
def test_compute_context(
    texts_key: str, expected: Tuple[int, int], kb_bert_predictor: TopKPredictor
) -> None:
    actual_start, actual_end = kb_bert_predictor.compute_context(TEXTS[texts_key])

    expected_start, expected_end = expected

    assert actual_start == expected_start
    assert actual_end == expected_end



from transformers import (  # type: ignore [import-untyped]
    FillMaskPipeline,
)

SCORE_FORMATS = {
    1: ("{:.1f}", lambda s: s.endswith(".0")),
    2: ("{:.2f}", lambda s: s.endswith(".00")),
    3: ("{:.3f}", lambda s: s.endswith(".000")),
    4: ("{:.4f}", lambda s: s.endswith(".0000")),
    5: ("{:.5f}", lambda s: s.endswith(".00000")),
    6: ("{:.6f}", lambda s: s.endswith(".000000")),
    7: ("{:.7f}", lambda s: s.endswith(".0000000")),
    8: ("{:.8f}", lambda s: s.endswith(".00000000")),
    9: ("{:.9f}", lambda s: s.endswith(".000000000")),
    10: ("{:.10f}", lambda s: s.endswith(".0000000000")),
}


class TopKPredictor:
    def __init__(self, *, tokenizer, model, num_decimals: int = 3) -> None:
        self.tokenizer = tokenizer
        self.model = model
        self.num_decimals = num_decimals
        self.pipeline = FillMaskPipeline(model=model, tokenizer=tokenizer)

    def get_top_k_predictions(self, text: str, k: int = 5) -> str:
        tokenized_inputs = self.tokenizer(text)
        if len(tokenized_inputs["input_ids"]) <= 512:
            return self._run_pipeline(text, k)
        if text.count("[MASK]") == 1:
            return self._run_pipeline_on_mask_context(text, k)
        raise RuntimeError(
            f"can't handle large input and multiple [MASK]: {len(tokenized_inputs['input_ids'])} tokens > 512 tokens"
        )

    def _run_pipeline_on_mask_context(self, text, k):
        start, end = self.compute_context(text)
        text_with_mask = text[start:end]
        return self._run_pipeline(text_with_mask, k)

    def compute_context(self, text):
        mask = text.find("[MASK]")
        lower = text[(mask - 210) : (mask - 190)].find(" ")
        higher = text[(mask + 190) : (mask + 210)].find(" ")
        start = mask - 210 + lower
        end = mask + 190 + higher
        return max(start, 0), min(end, len(text))

    def _run_pipeline(self, text, k) -> str:
        if predictions := self.pipeline(text, top_k=k):
            collect_token_and_score = (
                (pred["token_str"], pred["score"])  # type: ignore
                for pred in predictions
            )
            score_format, score_pred = SCORE_FORMATS[self.num_decimals]
            format_scores = (
                (token, score_format.format(score))
                for token, score in collect_token_and_score
            )
            filter_out_zero_scores = (
                (token, score)
                for token, score in format_scores
                if not score_pred(score)
            )
            predictions_str = "|".join(
                f"{token}:{score}" for token, score in filter_out_zero_scores
            )

            return f"|{predictions_str}|" if predictions_str else "|"
        else:
            return "|"



from dataclasses import dataclass
from typing import Optional, Tuple
from sparv.api import (  # type: ignore [import-untyped]
    annotator,
    Output,
    get_logger,
    Annotation,
    Config,
    SparvErrorMessage,
)

from transformers import (  # type: ignore [import-untyped]
    BertTokenizer,
    BertForMaskedLM,
)
from sbx_word_prediction_kb_bert.predictor import TopKPredictor

__description__ = "Calculating word predictions by mask a word in a BERT model."


__config__ = [
    Config(
        "sbx_word_prediction_kb_bert.num_predictions",
        description="The number of predictions to list",
        default=5,
    ),
    Config(
        "sbx_word_prediction_kb_bert.num_decimals",
        description="The number of decimals to round the score to",
        default=3,
    ),
]

__version__ = "0.5.2"

logger = get_logger(__name__)

TOK_SEP = " "


@dataclass
class HuggingfaceModel:
    model_name: str
    model_revision: str
    tokenizer_name: Optional[str] = None
    tokenizer_revision: Optional[str] = None

    def tokenizer_name_and_revision(self) -> Tuple[str, str]:
        if tokenizer_name := self.tokenizer_name:
            return tokenizer_name, self.tokenizer_revision or "main"
        else:
            return self.model_name, self.model_revision


MODELS = {
    "kb-bert": HuggingfaceModel(
        model_name="KBLab/bert-base-swedish-cased",
        model_revision="c710fb8dff81abb11d704cd46a8a1e010b2b022c",
    )
}


@annotator("Word prediction tagging with a masked Bert model", language=["swe"])
def predict_words__kb_bert(
    out_prediction: Output = Output(
        "<token>:sbx_word_prediction_kb_bert.word-prediction--kb-bert",
        cls="word_prediction",
        description="Word predictions from masked BERT (format: '|<word>:<score>|...|)",
    ),
    word: Annotation = Annotation("<token:word>"),
    sentence: Annotation = Annotation("<sentence>"),
    num_predictions_str: str = Config("sbx_word_prediction_kb_bert.num_predictions"),
    num_decimals_str: str = Config("sbx_word_prediction_kb_bert.num_decimals"),
) -> None:
    logger.info("predict_words")
    try:
        num_predictions = int(num_predictions_str)
    except ValueError as exc:
        raise SparvErrorMessage(
            f"'sbx_word_prediction_kb_bert.num_predictions' must contain an 'int' got: '{num_predictions_str}'"
        ) from exc
    try:
        num_decimals = int(num_decimals_str)
    except ValueError as exc:
        raise SparvErrorMessage(
            f"'sbx_word_prediction_kb_bert.num_decimals' must contain an 'int' got: '{num_decimals_str}'"
        ) from exc
    tokenizer_name, tokenizer_revision = MODELS["kb-bert"].tokenizer_name_and_revision()

    tokenizer = BertTokenizer.from_pretrained(
        tokenizer_name, revision=tokenizer_revision
    )
    model = BertForMaskedLM.from_pretrained(
        MODELS["kb-bert"].model_name, revision=MODELS["kb-bert"].model_revision
    )

    predictor = TopKPredictor(
        model=model,
        tokenizer=tokenizer,
        num_decimals=num_decimals,
    )

    sentences, _orphans = sentence.get_children(word)
    token_word = list(word.read())
    out_prediction_annotation = word.create_empty_attribute()

    run_word_prediction(
        predictor=predictor,
        num_predictions=num_predictions,
        sentences=sentences,
        token_word=token_word,
        out_prediction_annotations=out_prediction_annotation,
    )

    logger.info("writing annotations")
    out_prediction.write(out_prediction_annotation)


def run_word_prediction(
    predictor: TopKPredictor,
    num_predictions: int,
    sentences,
    token_word: list,
    out_prediction_annotations,
) -> None:
    logger.info("run_word_prediction")

    logger.progress(total=len(sentences))  # type: ignore
    for sent in sentences:
        logger.progress()  # type: ignore
        token_indices = list(sent)
        for token_index_to_mask in token_indices:
            sent_to_tag = TOK_SEP.join(
                (
                    "[MASK]"
                    if token_index == token_index_to_mask
                    else token_word[token_index]
                )
                for token_index in sent
            )

            predictions_scores = predictor.get_top_k_predictions(
                sent_to_tag, k=num_predictions
            )
            out_prediction_annotations[token_index_to_mask] = predictions_scores
